From 0f0e695510aeaee7ca24cb7ecaffa74f39bf9e67 Mon Sep 17 00:00:00 2001
From: Mike Palmiotto <mike.palmiotto@crunchydata.com>
Date: Wed, 11 Mar 2020 02:58:08 +0000
Subject: [PATCH 07/11] Add SysLoggerType to subprocess struct

This commit introduces two new fields to the subprocess struct:

1) keep_postmaster_memcontext allows subprocesses to specify whether or
not the PostmasterMemoryContext should be kept around after a fork.

2) postmaster_main allows postmaster to gain control to do some
necessary processing after a successful fork.
---
 src/backend/postmaster/postmaster.c |  49 +++--
 src/backend/postmaster/subprocess.c |  61 +++---
 src/backend/postmaster/syslogger.c  | 278 +++++++++++-----------------
 src/backend/utils/error/elog.c      |   4 +-
 src/include/postmaster/subprocess.h |  18 +-
 src/include/postmaster/syslogger.h  |  10 +-
 6 files changed, 199 insertions(+), 221 deletions(-)

diff --git a/src/backend/postmaster/postmaster.c b/src/backend/postmaster/postmaster.c
index d2e9301a1e..94acaaf375 100644
--- a/src/backend/postmaster/postmaster.c
+++ b/src/backend/postmaster/postmaster.c
@@ -355,6 +355,8 @@ bool		ClientAuthInProgress = false;	/* T during new-client
 											 * authentication */
 
 bool		redirection_done = false;	/* stderr redirected for syslogger? */
+LogFileNo	syslogFileNo = -1;
+LogFileNo	csvlogFileNo = -1;
 
 /* received START_AUTOVAC_LAUNCHER signal */
 static volatile sig_atomic_t start_autovac_launcher = false;
@@ -513,6 +515,10 @@ typedef struct
 	SubprocessType		MySubprocessType;
 
 	bool		redirection_done;
+#ifdef EXEC_BACKEND
+	LogFileNo	syslogFileNo;
+	LogFileNo	csvlogFileNo;
+#endif
 	bool		IsBinaryUpgrade;
 	int			max_safe_fds;
 	int			MaxBackends;
@@ -1083,7 +1089,7 @@ PostmasterMain(int argc, char *argv[])
 	/*
 	 * If enabled, start up syslogger collection subprocess
 	 */
-	SysLoggerPID = SysLogger_Start();
+	SysLoggerPID = StartSubprocess(SysLoggerType);
 
 	/*
 	 * Reset whereToSendOutput from DestDebug (its starting state) to
@@ -1739,7 +1745,7 @@ ServerLoop(void)
 
 		/* If we have lost the log collector, try to start a new one */
 		if (SysLoggerPID == 0 && Logging_collector)
-			SysLoggerPID = SysLogger_Start();
+			SysLoggerPID = StartSubprocess(SysLoggerType);
 
 		/*
 		 * If no background writer process is running, and we are not in a
@@ -3228,7 +3234,7 @@ reaper(SIGNAL_ARGS)
 		{
 			SysLoggerPID = 0;
 			/* for safety's sake, launch new logger *first* */
-			SysLoggerPID = SysLogger_Start();
+			SysLoggerPID = StartSubprocess(SysLoggerType);
 			if (!EXIT_STATUS_0(exitstatus))
 				LogChildExit(LOG, _("system logger process"),
 							 pid, exitstatus);
@@ -5063,12 +5069,6 @@ SubPostmasterMain(int argc, char *argv[])
 
 		StartBackgroundWorker();
 	}
-	if (strcmp(argv[1], "--forklog") == 0)
-	{
-		/* Do not want to attach to shared memory */
-
-		SysLoggerMain(argc, argv);	/* does not return */
-	}
 
 	if (MySubprocess->needs_aux_proc)
 	{
@@ -5446,6 +5446,15 @@ StartSubprocess(SubprocessType type)
 	snprintf(typebuf, sizeof(typebuf), "-x%d", type);
 	argv[argc++] = typebuf;
 
+	/* Prep subprocess for forking */
+	if (MySubprocess->fork_prep)
+	{
+		if (MySubprocess->fork_prep(argc, argv))
+		{
+			return 0;
+		}
+	}
+
 	argv[argc] = NULL;
 	Assert(argc < lengthof(argv));
 
@@ -5461,10 +5470,16 @@ StartSubprocess(SubprocessType type)
 		/* Close the postmaster's sockets */
 		ClosePostmasterPorts(false);
 
-		/* Release postmaster's working memory context */
-		MemoryContextSwitchTo(TopMemoryContext);
-		MemoryContextDelete(PostmasterContext);
-		PostmasterContext = NULL;
+		/*
+		 * Release postmaster's working memory context if the subprocess doesn't
+		 * need it
+		 */
+		if (!MySubprocess->keep_postmaster_memcontext)
+		{
+			MemoryContextSwitchTo(TopMemoryContext);
+			MemoryContextDelete(PostmasterContext);
+			PostmasterContext = NULL;
+		}
 
 		/*
 		 * Drop connection to postmaster's shared memory if the subprocess
@@ -5508,6 +5523,10 @@ StartSubprocess(SubprocessType type)
 		return 0;
 	}
 
+	/* Some subprocesses need to do more work in postmaster. */
+	if (MySubprocess->postmaster_main)
+		MySubprocess->postmaster_main(argc, argv);
+
 	/*
 	 * in parent, successful fork
 	 */
@@ -6185,6 +6204,8 @@ save_backend_variables(BackendParameters *param, Port *port,
 	param->MySubprocessType = MySubprocessType;
 
 	param->redirection_done = redirection_done;
+	param->syslogFileNo = syslogFileNo;
+	param->csvlogFileNo = csvlogFileNo;
 	param->IsBinaryUpgrade = IsBinaryUpgrade;
 	param->max_safe_fds = max_safe_fds;
 
@@ -6422,6 +6443,8 @@ restore_backend_variables(BackendParameters *param, Port *port)
 	MySubprocessType = param->MySubprocessType;
 
 	redirection_done = param->redirection_done;
+	syslogFileNo = param->syslogFileNo;
+	csvlogFileNo = param->csvlogFileNo;
 	IsBinaryUpgrade = param->IsBinaryUpgrade;
 	max_safe_fds = param->max_safe_fds;
 
diff --git a/src/backend/postmaster/subprocess.c b/src/backend/postmaster/subprocess.c
index f9d71a627c..122eb31b41 100644
--- a/src/backend/postmaster/subprocess.c
+++ b/src/backend/postmaster/subprocess.c
@@ -19,6 +19,7 @@
 #include "postmaster/postmaster.h"
 #include "postmaster/startup.h"
 #include "postmaster/subprocess.h"
+#include "postmaster/syslogger.h"
 #include "postmaster/walwriter.h"
 #include "replication/walreceiver.h"
 
@@ -31,25 +32,23 @@ static PgSubprocess process_types[] = {
 		.desc = "checker",
 		.needs_aux_proc = true,
 		.needs_shmem = true,
-		.fork_prep = NULL,
-		.entrypoint = CheckerModeMain,
-		.fork_failure = NULL
+		.keep_postmaster_memcontext = false,
+		.entrypoint = CheckerModeMain
 	},
 	{
 		.name = "boot",
 		.desc = "bootstrap",
 		.needs_aux_proc = true,
 		.needs_shmem = true,
-		.fork_prep = NULL,
-		.entrypoint = BootstrapModeMain,
-		.fork_failure = NULL
+		.keep_postmaster_memcontext = false,
+		.entrypoint = BootstrapModeMain
 	},
 	{
 		.name = "boot",
 		.desc = "startup",
 		.needs_aux_proc = true,
 		.needs_shmem = true,
-		.fork_prep = NULL,
+		.keep_postmaster_memcontext = false,
 		.entrypoint = StartupProcessMain,
 		.fork_failure = StartupForkFailure
 	},
@@ -58,72 +57,76 @@ static PgSubprocess process_types[] = {
 		.desc = "background writer",
 		.needs_aux_proc = true,
 		.needs_shmem = true,
-		.fork_prep = NULL,
-		.entrypoint = BackgroundWriterMain,
-		.fork_failure = NULL
+		.keep_postmaster_memcontext = false,
+		.entrypoint = BackgroundWriterMain
 	},
 	{
 		.name = "boot",
 		.desc = "checkpointer",
 		.needs_aux_proc = true,
 		.needs_shmem = true,
-		.fork_prep = NULL,
-		.entrypoint = CheckpointerMain,
-		.fork_failure = NULL
+		.keep_postmaster_memcontext = false,
+		.entrypoint = CheckpointerMain
 	},
 	{
 		.name = "boot",
 		.desc = "wal writer",
 		.needs_aux_proc = true,
 		.needs_shmem = true,
-		.fork_prep = NULL,
-		.entrypoint = WalWriterMain,
-		.fork_failure = NULL
+		.keep_postmaster_memcontext = false,
+		.entrypoint = WalWriterMain
 	},
 	{
 		.name = "boot",
 		.desc = "wal receiver",
 		.needs_aux_proc = true,
 		.needs_shmem = true,
-		.fork_prep = NULL,
-		.entrypoint = WalReceiverMain,
-		.fork_failure = NULL
+		.keep_postmaster_memcontext = false,
+		.entrypoint = WalReceiverMain
 	},
 	{
 		.name = "avlauncher",
 		.desc = "autovacuum launcher",
 		.needs_aux_proc = false,
 		.needs_shmem = true,
-		.fork_prep = NULL,
-		.entrypoint = AutoVacLauncherMain,
-		.fork_failure = NULL
+		.keep_postmaster_memcontext = false,
+		.entrypoint = AutoVacLauncherMain
 	},
 	{
 		.name = "avworker",
 		.desc = "autovacuum worker",
 		.needs_aux_proc = false,
 		.needs_shmem = true,
-		.fork_prep = NULL,
-		.entrypoint = AutoVacWorkerMain,
-		.fork_failure = NULL
+		.keep_postmaster_memcontext = false,
+		.entrypoint = AutoVacWorkerMain
 	},
 	{
 		.name = "col",
 		.desc = "statistics collector",
 		.needs_aux_proc = false,
 		.needs_shmem = false,
+		.keep_postmaster_memcontext = false,
 		.fork_prep = PgstatCollectorPrep,
-		.entrypoint = PgstatCollectorMain,
-		.fork_failure = NULL
+		.entrypoint = PgstatCollectorMain
 	},
 	{
 		.name = "arch",
 		.desc = "archiver",
 		.needs_aux_proc = false,
 		.needs_shmem = false,
+		.keep_postmaster_memcontext = false,
 		.fork_prep = PgArchiverPrep,
-		.entrypoint = PgArchiverMain,
-		.fork_failure = NULL
+		.entrypoint = PgArchiverMain
+	},
+	{
+		.name = "log",
+		.desc = "syslogger",
+		.needs_aux_proc = false,
+		.needs_shmem = false,
+		.keep_postmaster_memcontext = true,
+		.fork_prep = SysLoggerPrep,
+		.entrypoint = SysLoggerMain,
+		.postmaster_main = SysLoggerPostmasterMain
 	}
 };
 
diff --git a/src/backend/postmaster/syslogger.c b/src/backend/postmaster/syslogger.c
index ffcb54968f..8824196c43 100644
--- a/src/backend/postmaster/syslogger.c
+++ b/src/backend/postmaster/syslogger.c
@@ -38,8 +38,8 @@
 #include "nodes/pg_list.h"
 #include "pgstat.h"
 #include "pgtime.h"
-#include "postmaster/fork_process.h"
 #include "postmaster/postmaster.h"
+#include "postmaster/subprocess.h"
 #include "postmaster/syslogger.h"
 #include "storage/dsm.h"
 #include "storage/fd.h"
@@ -75,6 +75,10 @@ bool		Log_truncate_on_rotation = false;
 int			Log_file_mode = S_IRUSR | S_IWUSR;
 
 extern bool redirection_done;
+#ifdef EXEC_BACKEND
+extern LogFileNo	syslogFileNo;
+extern LogFileNo	csvlogFileNo;
+#endif
 
 /*
  * Private state
@@ -129,10 +133,8 @@ static volatile sig_atomic_t rotation_requested = false;
 
 /* Local subroutines */
 #ifdef EXEC_BACKEND
-static pid_t syslogger_forkexec(void);
-static void syslogger_parseArgs(int argc, char *argv[]);
+static void syslogger_openfds(void);
 #endif
-NON_EXEC_STATIC void SysLoggerMain(int argc, char *argv[]) pg_attribute_noreturn();
 static void process_pipe_input(char *logbuffer, int *bytes_in_logbuffer);
 static void flush_pipe_input(char *logbuffer, int *bytes_in_logbuffer);
 static FILE *logfile_open(const char *filename, const char *mode,
@@ -153,7 +155,7 @@ static void update_metainfo_datafile(void);
  * Main entry point for syslogger process
  * argc/argv parameters are valid only in EXEC_BACKEND case.
  */
-NON_EXEC_STATIC void
+void
 SysLoggerMain(int argc, char *argv[])
 {
 #ifndef WIN32
@@ -169,10 +171,9 @@ SysLoggerMain(int argc, char *argv[])
 	now = MyStartTime;
 
 #ifdef EXEC_BACKEND
-	syslogger_parseArgs(argc, argv);
+	syslogger_openfds();
 #endif							/* EXEC_BACKEND */
 
-	MyBackendType = B_LOGGER;
 	init_ps_display(NULL);
 
 	/*
@@ -537,13 +538,13 @@ SysLoggerMain(int argc, char *argv[])
  * Postmaster subroutine to start a syslogger subprocess.
  */
 int
-SysLogger_Start(void)
+SysLoggerPrep(int argc, char *argv[])
 {
-	pid_t		sysloggerPid;
 	char	   *filename;
 
+	/* Tell postmaster to stop forking */
 	if (!Logging_collector)
-		return 0;
+		return -1;
 
 	/*
 	 * If first time through, create the pipe which will receive stderr
@@ -626,179 +627,124 @@ SysLogger_Start(void)
 	}
 
 #ifdef EXEC_BACKEND
-	switch ((sysloggerPid = syslogger_forkexec()))
-#else
-	switch ((sysloggerPid = fork_process()))
-#endif
-	{
-		case -1:
-			ereport(LOG,
-					(errmsg("could not fork system logger: %m")));
-			return 0;
-
-#ifndef EXEC_BACKEND
-		case 0:
-			/* in postmaster child ... */
-			InitPostmasterChild();
-
-			/* Close the postmaster's sockets */
-			ClosePostmasterPorts(true);
-
-			/* Drop our connection to postmaster's shared memory, as well */
-			dsm_detach_all();
-			PGSharedMemoryDetach();
-
-			/* do the work */
-			SysLoggerMain(0, NULL);
-			break;
-#endif
-
-		default:
-			/* success, in postmaster */
-
-			/* now we redirect stderr, if not done already */
-			if (!redirection_done)
-			{
-#ifdef WIN32
-				int			fd;
-#endif
-
-				/*
-				 * Leave a breadcrumb trail when redirecting, in case the user
-				 * forgets that redirection is active and looks only at the
-				 * original stderr target file.
-				 */
-				ereport(LOG,
-						(errmsg("redirecting log output to logging collector process"),
-						 errhint("Future log output will appear in directory \"%s\".",
-								 Log_directory)));
-
-#ifndef WIN32
-				fflush(stdout);
-				if (dup2(syslogPipe[1], fileno(stdout)) < 0)
-					ereport(FATAL,
-							(errcode_for_file_access(),
-							 errmsg("could not redirect stdout: %m")));
-				fflush(stderr);
-				if (dup2(syslogPipe[1], fileno(stderr)) < 0)
-					ereport(FATAL,
-							(errcode_for_file_access(),
-							 errmsg("could not redirect stderr: %m")));
-				/* Now we are done with the write end of the pipe. */
-				close(syslogPipe[1]);
-				syslogPipe[1] = -1;
-#else
-
-				/*
-				 * open the pipe in binary mode and make sure stderr is binary
-				 * after it's been dup'ed into, to avoid disturbing the pipe
-				 * chunking protocol.
-				 */
-				fflush(stderr);
-				fd = _open_osfhandle((intptr_t) syslogPipe[1],
-									 _O_APPEND | _O_BINARY);
-				if (dup2(fd, _fileno(stderr)) < 0)
-					ereport(FATAL,
-							(errcode_for_file_access(),
-							 errmsg("could not redirect stderr: %m")));
-				close(fd);
-				_setmode(_fileno(stderr), _O_BINARY);
-
-				/*
-				 * Now we are done with the write end of the pipe.
-				 * CloseHandle() must not be called because the preceding
-				 * close() closes the underlying handle.
-				 */
-				syslogPipe[1] = 0;
-#endif
-				redirection_done = true;
-			}
-
-			/* postmaster will never write the file(s); close 'em */
-			fclose(syslogFile);
-			syslogFile = NULL;
-			if (csvlogFile != NULL)
-			{
-				fclose(csvlogFile);
-				csvlogFile = NULL;
-			}
-			return (int) sysloggerPid;
-	}
-
-	/* we should never reach here */
-	return 0;
-}
-
-
-#ifdef EXEC_BACKEND
-
-/*
- * syslogger_forkexec() -
- *
- * Format up the arglist for, then fork and exec, a syslogger process
- */
-static pid_t
-syslogger_forkexec(void)
-{
-	char	   *av[10];
-	int			ac = 0;
-	char		filenobuf[32];
-	char		csvfilenobuf[32];
-
-	av[ac++] = "postgres";
-	av[ac++] = "--forklog";
-	av[ac++] = NULL;			/* filled in by postmaster_forkexec */
-
-	/* static variables (those not passed by write_backend_variables) */
 #ifndef WIN32
 	if (syslogFile != NULL)
-		snprintf(filenobuf, sizeof(filenobuf), "%d",
-				 fileno(syslogFile));
+		syslogFileNo = fileno(syslogFile);
 	else
-		strcpy(filenobuf, "-1");
+		syslogFileNo = -1;
 #else							/* WIN32 */
 	if (syslogFile != NULL)
-		snprintf(filenobuf, sizeof(filenobuf), "%ld",
-				 (long) _get_osfhandle(_fileno(syslogFile)));
+		syslogFileNo = (long) _get_osfhandle(_fileno(syslogFile));
 	else
-		strcpy(filenobuf, "0");
+		syslogFileNo = 0;
 #endif							/* WIN32 */
-	av[ac++] = filenobuf;
 
 #ifndef WIN32
 	if (csvlogFile != NULL)
-		snprintf(csvfilenobuf, sizeof(csvfilenobuf), "%d",
-				 fileno(csvlogFile));
+		csvlogFileNo = fileno(csvlogFile);
 	else
-		strcpy(csvfilenobuf, "-1");
+		csvlogFileNo = -1;
 #else							/* WIN32 */
 	if (csvlogFile != NULL)
-		snprintf(csvfilenobuf, sizeof(csvfilenobuf), "%ld",
-				 (long) _get_osfhandle(_fileno(csvlogFile)));
+		csvlogFileNo = (long) _get_osfhandle(_fileno(csvlogFile));
 	else
-		strcpy(csvfilenobuf, "0");
+		csvlogFileNo = 0;
 #endif							/* WIN32 */
-	av[ac++] = csvfilenobuf;
+#endif							/* EXEC_BACKEND */
+
+	return 0;
+}
+
+/*
+ * SysLoggerPostmasterMain
+ *
+ * Control function for subprocess parent (postmaster) after a successful fork.
+ */
+void
+SysLoggerPostmasterMain(int argc, char *argv[])
+{
+	/* success, in postmaster */
 
-	av[ac] = NULL;
-	Assert(ac < lengthof(av));
+	/* now we redirect stderr, if not done already */
+	if (!redirection_done)
+	{
+#ifdef WIN32
+		int			fd;
+#endif
+
+		/*
+		 * Leave a breadcrumb trail when redirecting, in case the user
+		 * forgets that redirection is active and looks only at the
+		 * original stderr target file.
+		 */
+		ereport(LOG,
+				(errmsg("redirecting log output to logging collector process"),
+				 errhint("Future log output will appear in directory \"%s\".",
+						 Log_directory)));
 
-	return postmaster_forkexec(ac, av);
+#ifndef WIN32
+		fflush(stdout);
+		if (dup2(syslogPipe[1], fileno(stdout)) < 0)
+			ereport(FATAL,
+					(errcode_for_file_access(),
+					 errmsg("could not redirect stdout: %m")));
+		fflush(stderr);
+		if (dup2(syslogPipe[1], fileno(stderr)) < 0)
+			ereport(FATAL,
+					(errcode_for_file_access(),
+					 errmsg("could not redirect stderr: %m")));
+		/* Now we are done with the write end of the pipe. */
+		close(syslogPipe[1]);
+		syslogPipe[1] = -1;
+#else
+
+		/*
+		 * open the pipe in binary mode and make sure stderr is binary
+		 * after it's been dup'ed into, to avoid disturbing the pipe
+		 * chunking protocol.
+		 */
+		fflush(stderr);
+		fd = _open_osfhandle((intptr_t) syslogPipe[1],
+							 _O_APPEND | _O_BINARY);
+		if (dup2(fd, _fileno(stderr)) < 0)
+			ereport(FATAL,
+					(errcode_for_file_access(),
+					 errmsg("could not redirect stderr: %m")));
+		close(fd);
+		_setmode(_fileno(stderr), _O_BINARY);
+
+		/*
+		 * Now we are done with the write end of the pipe.
+		 * CloseHandle() must not be called because the preceding
+		 * close() closes the underlying handle.
+		 */
+		syslogPipe[1] = 0;
+#endif
+		redirection_done = true;
+	}
+
+	/* postmaster will never write the file(s); close 'em */
+	fclose(syslogFile);
+	syslogFile = NULL;
+	if (csvlogFile != NULL)
+	{
+		fclose(csvlogFile);
+		csvlogFile = NULL;
+	}
 }
 
+#ifdef EXEC_BACKEND
 /*
- * syslogger_parseArgs() -
+ * syslogger_openfds() -
  *
  * Extract data from the arglist for exec'ed syslogger process
  */
 static void
-syslogger_parseArgs(int argc, char *argv[])
+syslogger_openfds(void)
 {
-	int			fd;
-
-	Assert(argc == 5);
-	argv += 3;
-
+#ifdef WIN32
+	int			fd = 0;
+#endif
 	/*
 	 * Re-open the error output files that were opened by SysLogger_Start().
 	 *
@@ -807,21 +753,18 @@ syslogger_parseArgs(int argc, char *argv[])
 	 * coded, we'll just crash on a null pointer dereference after failure...
 	 */
 #ifndef WIN32
-	fd = atoi(*argv++);
-	if (fd != -1)
+	if (syslogFileNo != -1)
 	{
-		syslogFile = fdopen(fd, "a");
+		syslogFile = fdopen(syslogFileNo, "a");
 		setvbuf(syslogFile, NULL, PG_IOLBF, 0);
 	}
-	fd = atoi(*argv++);
-	if (fd != -1)
+	if (csvlogFileNo != -1)
 	{
-		csvlogFile = fdopen(fd, "a");
+		csvlogFile = fdopen(csvlogFileNo, "a");
 		setvbuf(csvlogFile, NULL, PG_IOLBF, 0);
 	}
 #else							/* WIN32 */
-	fd = atoi(*argv++);
-	if (fd != 0)
+	if (syslogFileNo != 0)
 	{
 		fd = _open_osfhandle(fd, _O_APPEND | _O_TEXT);
 		if (fd > 0)
@@ -830,8 +773,7 @@ syslogger_parseArgs(int argc, char *argv[])
 			setvbuf(syslogFile, NULL, PG_IOLBF, 0);
 		}
 	}
-	fd = atoi(*argv++);
-	if (fd != 0)
+	if (syslogFileNo != 0)
 	{
 		fd = _open_osfhandle(fd, _O_APPEND | _O_TEXT);
 		if (fd > 0)
diff --git a/src/backend/utils/error/elog.c b/src/backend/utils/error/elog.c
index b8858b132b..4ec598a28d 100644
--- a/src/backend/utils/error/elog.c
+++ b/src/backend/utils/error/elog.c
@@ -2839,7 +2839,7 @@ write_csvlog(ErrorData *edata)
 	appendStringInfoChar(&buf, '\n');
 
 	/* If in the syslogger process, try to write messages direct to file */
-	if (MyBackendType == B_LOGGER)
+	if (MyBackendType == B_LOGGER && IsUnderPostmaster)
 		write_syslogger_file(buf.data, buf.len, LOG_DESTINATION_CSVLOG);
 	else
 		write_pipe_chunks(buf.data, buf.len, LOG_DESTINATION_CSVLOG);
@@ -3052,7 +3052,7 @@ send_message_to_server_log(ErrorData *edata)
 	}
 
 	/* If in the syslogger process, try to write messages direct to file */
-	if (MyBackendType == B_LOGGER)
+	if (MyBackendType == B_LOGGER && IsUnderPostmaster)
 		write_syslogger_file(buf.data, buf.len, LOG_DESTINATION_STDERR);
 
 	/* Write to CSV log if enabled */
diff --git a/src/include/postmaster/subprocess.h b/src/include/postmaster/subprocess.h
index 3f00f904dc..d0d98a5d69 100644
--- a/src/include/postmaster/subprocess.h
+++ b/src/include/postmaster/subprocess.h
@@ -27,6 +27,7 @@ typedef enum
 	AutoVacuumWorkerType,
 	PgstatCollectorType,
 	PgArchiverType,
+	SysLoggerType,
 
 	NUMSUBPROCESSTYPES			/* Must be last! */
 } SubprocessType;
@@ -34,19 +35,22 @@ typedef enum
 typedef int		(*SubprocessPrep) (int argc, char *argv[]);
 typedef void	(*SubprocessEntryPoint) (int argc, char *argv[]);
 typedef bool	(*SubprocessForkFailure) (int fork_errno);
+typedef void	(*SubprocessPostmasterMain) (int argc, char *argv[]);
 
 /* Current subprocess initializer */
 extern void InitMySubprocess(SubprocessType type);
 
 typedef struct PgSubprocess
 {
-	const char			   *name;
-	const char			   *desc;
-	bool					needs_aux_proc;
-	bool					needs_shmem;
-	SubprocessPrep			fork_prep;
-	SubprocessEntryPoint	entrypoint;
-	SubprocessForkFailure	fork_failure;
+	const char				   *name;
+	const char				   *desc;
+	bool						needs_aux_proc;
+	bool						needs_shmem;
+	bool						keep_postmaster_memcontext;
+	SubprocessPrep				fork_prep;
+	SubprocessEntryPoint		entrypoint;
+	SubprocessForkFailure		fork_failure;
+	SubprocessPostmasterMain	postmaster_main;
 } PgSubprocess;
 
 extern SubprocessType				MySubprocessType;
diff --git a/src/include/postmaster/syslogger.h b/src/include/postmaster/syslogger.h
index f611bd1411..6fbcb0c509 100644
--- a/src/include/postmaster/syslogger.h
+++ b/src/include/postmaster/syslogger.h
@@ -76,14 +76,20 @@ extern int	syslogPipe[2];
 extern HANDLE syslogPipe[2];
 #endif
 
+#ifndef WIN32
+typedef int		LogFileNo;
+#else
+typedef long	LogFileNo;
+#endif
 
 extern int	SysLogger_Start(void);
 
 extern void write_syslogger_file(const char *buffer, int count, int dest);
 
-#ifdef EXEC_BACKEND
+/* Subprocess startup functions */
+extern int SysLoggerPrep(int argc, char *argv[]);
 extern void SysLoggerMain(int argc, char *argv[]) pg_attribute_noreturn();
-#endif
+extern void SysLoggerPostmasterMain(int argc, char *argv[]);
 
 extern bool CheckLogrotateSignal(void);
 extern void RemoveLogrotateSignalFiles(void);
-- 
2.21.1

